#include <stdio.h>
#include <math.h>
#include <string.h>
#include "fr_recon.h"
#include "fr_paramest.h"
#include "fr_icm.h"
#include "../ansi_lib_loc/io_mat_dbl.h"
#include "../ansi_lib_loc/allocate.h"
#include "../ansi_lib_loc/assort.h"
#include "../ansi_lib_loc/multigrid.h"
#include "../ansi_lib_loc/solve2.h"
#include "../ansi_lib_loc/randlib.h"
#include "../ansi_lib_loc/A_comp_dbl_storeflt.h"
#ifdef __linux__
#include <getopt.h>
#endif
extern int optind;
extern char *optarg;

/*#define RECON_MASK*/

void cmdline_error(void);
void parse_cmdline(struct cmdline_params *cmdline,int argc,char **argv);
void read_data(struct all_data *data,struct column *sp_col,struct cmdline_params *cmdline);
void write_output(struct cmdline_params *cmdline,struct all_data *data,double **sigma,struct geom_info *geom);
void free_memory(struct all_data *data, struct column *sp_col,double **sigma,struct cmdline_params *cmdline);
void init_sp_col(struct column *sp_col,struct sino_info *sino);
double recon(struct all_data *data,struct geom_info *geom,struct column *sp_col,struct cmdline_params *cmdline);
double calc_init_const(struct all_data *data,struct geom_info *geom,double **profiles,struct column *sp_col);
void init_X_to_constant(struct all_data *data,struct geom_info *geom,double **profiles,struct column *sp_col);
void calculate_pix_recon_mask(struct all_data *data,struct geom_info *geom);
void clip_initial(struct all_data *data,struct geom_info *geom);




int main( int argc, char **argv )
{
  int N,M,resol;
  struct geom_info geom;
  struct column sp_col;
  struct cmdline_params cmdline;
  struct all_data data;
  double **sigma;

  /* parse command-line */
  parse_cmdline(&cmdline,argc,argv);

  /* set A_correction-mode to emission */
  if (cmdline.Transcor_FLAG)
    sp_col.A_cor.mode = 'e'; /* attenuation correction ON */
  else
    sp_col.A_cor.mode = 't'; /* attenuation correction OFF */ 

  /* init random generator */
  srandom2(1);
  

  /* read input */
  read_data(&data,&sp_col,&cmdline);

  /* set scatter to minimum if zero */
  if(data.scatter == 0.0)
    data.scatter = 0.001;
  
  /* set reconstruction geometry */
  resol    = cmdline.resol;
  geom.dia = cmdline.fov;

  geom.n_y = resol;
  geom.n_x = resol;
  geom.delta = geom.dia/(double)resol;
  geom.x_0 = (-geom.dia+geom.delta)/2.0;
  geom.y_0 = (-geom.dia+geom.delta)/2.0;

  /* shortcuts for recon. image size */
  N=geom.n_y;
  M=geom.n_x;
  
  /* Allocate memory and initialize projection matrices*/
  init_sp_col(&sp_col,&data.sino);
  sigma = (double **)get_img(1,1,sizeof(double));
 
  /* calc_pix_recon_mask: if pix outside t_0 or outside fov,
     will set to zero */
  calculate_pix_recon_mask(&data,&geom);
  clip_initial(&data,&geom);

  /* fix-res reconstruction */
  sigma[0][0] = recon(&data,&geom,&sp_col,&cmdline);
  
  write_output(&cmdline,&data,sigma,&geom);

  free_memory(&data, &sp_col,sigma,&cmdline);

  return(0);

}




double recon(struct all_data *data,struct geom_info *geom,struct column *sp_col,struct cmdline_params *cmdline)
{
   
  double **profiles;
  int i;
  static double *S;
  static int param, N,M;
  
  static double SX=0.0,P;

  srandom2(1);

  P = cmdline->p-1;
  
  /* compute pixel profiles, data needed for computing A */
  fflush(stderr);
  printf("Calculating pixel profiles for system matrix...\n");
  fflush(stdout);
  pix_prof_comp(&data->sino,geom,&profiles);
  printf("done.\n\n");fflush(stdout);fflush(stderr);
 
  N = geom->n_y;
  M = geom->n_x;
 
  
  /* If constant initialization, calculate constant 
     such that total counts of forward projection
     matches sinogram data */
  if (!strcmp(cmdline->initial,"none"))
    init_X_to_constant(data,geom,profiles,sp_col);




  
  /* PARAMETER ESTIMATION */
  if( cmdline->no_iter_param != 0)
    {
      /* calculate number of pixels w/ nonzero mask  Nsup */
      data->Nsup=0;
      for(i=0;i<N*M;i++)
	if(data->mask[0][i])
	  data->Nsup++;
      
      SX=parameter_estimation(data, geom,profiles, cmdline ,sp_col, P);
    }

  
  SX = pow(cmdline->sigma,cmdline->p);
  printf("Sigma is %f\n",pow(SX,1.0/cmdline->p));


  /* reconstruct */
  if(cmdline->no_iter_rec != 0)
    {
      /* Allocate S  which is Ax*/
      S = (double *) get_spc(data->sino.n_t*data->sino.n_theta, sizeof(double));

      /* Calculate forward projection of initial image X, store in S */
      f_pro(S,data,geom,profiles,sp_col);


      printf("Starting reconstruction ...\n");
      param = 0;
      SX=relax_recon(data,S, geom, profiles, cmdline,sp_col,P,SX);


      printf("Finished reconstruction\n");
      free((void *)S);
    }
  
  
  free_img((void **)profiles);
  
  return(pow(SX,1.0/(P+1)));
  
}


void init_sp_col(struct column *sp_col,struct sino_info *sino)
{
  sp_col->val = (double *) get_spc(sino->n_t*sino->n_theta, sizeof(double));
  sp_col->row = (int *) get_spc(sino->n_t*sino->n_theta, sizeof(int));
  sp_col->multiplier = 1.0; 
}


void cmdline_error()
{
  fprintf(stderr,"Usage:\n");
  fprintf(stderr,"emisnew [-i #iter (10)] [-s sigma (10.0)] [-p GGMRF_param (2.0)] [-f field_of_view_mm (200.0)] [-r recon_resol (128)] [-I interpolate_output] [-T attn_corr.mat] input.mat initial.mat output.mat\n");
  fprintf(stderr,"\nUse 'none' for constant initial condition.\n\n");
  exit(-1); 
}




void parse_cmdline(struct cmdline_params *cmdline,int argc,char **argv)
{
  int ct=0;
  int i;

  /* set defaults */
  cmdline->no_iter_param=0;
  cmdline->p=2.0;
  cmdline->no_iter_rec=10;
  cmdline->sigma=10.0;
  cmdline->interpolate=0;
  cmdline->fov=200.0;
  cmdline->resol=128;
  cmdline->mask_FLAG=0;
  cmdline->Transcor_FLAG=0;

  for (;;)
    { 
      i= getopt(argc, argv, "i:s:p:If:r:E:m:T:");
      if (i==EOF)
        break;
      else
        {
          if (i == '?')
            cmdline_error();
	  if (i=='i')
	    {
	      cmdline->no_iter_rec=atoi(optarg);
	    }
	  if (i=='s')
	    {
	      cmdline->sigma=atof(optarg);
	    }
	  if (i=='p')
	    {
	      cmdline->p=atof(optarg);
	    }  
	  if (i=='I')
	    {
	      cmdline->interpolate=1;
	    }
	  if (i=='f')
	    {
	      cmdline->fov=atof(optarg);
	    }
	  if (i=='r')
	    {
	      cmdline->resol=atoi(optarg);
	    }
	  if (i=='E')
	    {
	      cmdline->no_iter_param=atoi(optarg);
	    }
	  if (i=='m')
	    {
	      strcpy(cmdline->mask_fname,optarg);  
	      cmdline->mask_FLAG=1;
	    }
	  if (i=='T')
	    {
	      strcpy(cmdline->attenuation,optarg);  
	      cmdline->Transcor_FLAG=1;
	    }
	  
	}
    }

  ct=optind;
  
  if ((argc-optind < 3) || (argc-optind > 3))
    {
      cmdline_error();
    }

  cmdline->input   = argv[ct];
  ct++;
  cmdline->initial = argv[ct];
  ct++;
  cmdline->output = argv[ct];
  ct++;


  if (( cmdline->no_iter_param>0) && (cmdline->mask_FLAG==0))
    {
      fprintf(stderr,"Input Error: parameter estimation swicted on (-E flag) but no mask specified (-m).\n\a\n");
      exit(-1);
    }


  printf("\nCommand-line arguments:\n");
  printf(" Input file : %s\n Initial    : %s\n Output file: %s\n",cmdline->input,cmdline->initial,cmdline->output);
  printf(" Parameters : p=%f, sigma=%f, no_iter_rec %d\n",cmdline->p,cmdline->sigma,cmdline->no_iter_rec);
  printf("              fov=%fmm  resol=%dpix\n",cmdline->fov,cmdline->resol);
  if (cmdline->Transcor_FLAG)
    printf(" Attenuation correction ON, trn sinogram file: %s\n",cmdline->attenuation);
  else
    printf(" NO attenuation correction\n");

  printf("\n\n");
}




void read_data(struct all_data *data,struct column *sp_col,struct cmdline_params *cmdline)
{
  MATFile *fin;   
  mxArray *pm;
  

  double **temp_image;
  int i;

  int N,M,Ang,T;

  
  /* read input sinogram Y */
  printf("Reading input sinogram, parameters:\n");
  fflush(stdout);fflush(stderr);
  fin = matOpen(cmdline->input, "r");
  if (fin == NULL)
    {
      fprintf(stderr,"Cannot read MatLab file %s\n",cmdline->input);
      exit(-1);
    }
  
  data->Y = read_mat_double_image(&Ang,&T,"counts",fin);
  if (data->Y == NULL)
    {
      fprintf(stderr,"Cannot read sino matrix \n");
      exit(-1);
    }
  
  
  read_mat_sino_info(fin, &data->sino);
  
  pm = matGetArray(fin,"scatter");
  if(pm != NULL) {
    data->scatter = mxGetScalar(pm);
  }

  
  matClose(fin);
  fflush(stdout);fflush(stderr);
  

  /* Read initial X */
  if (strcmp(cmdline->initial,"none"))
    {
      printf("Reading initial condition..");fflush(stdout);fflush(stderr);
      fin = matOpen(cmdline->initial, "r");
      if (fin == NULL)
	{
	  fprintf(stderr,"Cannot read MatLab file %s\n\a\n",cmdline->initial);
	  exit(-1);
	}
      
      data->X = read_mat_double_image(&N,&M,"img",fin);
      if (data->X == NULL)
	{
	  fprintf(stderr,"Cannot read img matrix \n\a\n");
	  exit(-1);
	}
      matClose(fin);
      if ((N!=cmdline->resol) || (M!=cmdline->resol))
	{
	  fprintf(stderr,"Error: Size of initial condition (%d by %d) doesn't match specified resolution of %d.\n\a\n",N,M,cmdline->resol);
	  exit(-1);
	}
      printf("done.\n");fflush(stdout);fflush(stderr);
    }
  else
    data->X = (double**)get_img(cmdline->resol,cmdline->resol,sizeof(double));
  


  if (cmdline->mask_FLAG)
    {
      /* read support = MASK for parameter estimation */
      fin = matOpen(cmdline->mask_fname, "r");
      if (fin == NULL)
	{ 
	  fprintf(stderr,"Cannot read MatLab file %s\n",cmdline->mask_fname);
	  exit(-1);
	} 
      
      temp_image = read_mat_double_image(&N,&M,"mask",fin);
      if (temp_image == NULL)
	{
	  fprintf(stderr,"Cannot read mask matrix \n");
	  exit(-1);
	}
      matClose(fin);
      
      /* convert double mask into integer mask */
      data->mask = (int **)get_img(M,N,sizeof(int)); 
      for(i=0;i<N*M;i++) 
	data->mask[0][i] = (int)temp_image[0][i];
      
      free_img((void **)temp_image); 
    }



  /* Read attenuation */
  if (sp_col->A_cor.mode == 'e')
    {                         
      fin = matOpen(cmdline->attenuation, "r");
      if (fin == NULL)
	{
	  fprintf(stderr,"Error: Cannot read attenuation file %s\n\a\n",cmdline->attenuation);
	  exit(-1);
	}
      else
	{ 
	  sp_col->A_cor.Y_t = read_mat_double_image(&Ang,&T,"trn",fin);
	  if (sp_col->A_cor.Y_t == NULL)
	    {
	      fprintf(stderr,"Cannot read trn matrix in file %s.\n\a\n",cmdline->attenuation);
	      exit(-1);
	    }
	  matClose(fin);
	} 
    } 
  
  

  printf("Done reading data.\n\n");
  
}
  
void write_output(struct cmdline_params *cmdline,struct all_data *data,double **sigma,struct geom_info *geom)
{
  MATFile *fout;
  int N,M,i,j;
  char matrix[20];
  double **tmp;
  double fov,*ptmp;

  N=geom->n_y;
  M=geom->n_x;
  printf("Writing output image %s ...\n",cmdline->output);
  
  
  fout = matOpen(cmdline->output, "w");
  if (fout == NULL)
    {
      fprintf(stderr,"Cannot write MatLab file %s\n",cmdline->output);
      exit(-1);
    }
      

  sprintf(matrix,"img");
  if (cmdline->interpolate==0)
    {
      write_mat_double_image(data->X,N,M,matrix,fout);
    }
  else
    {
      tmp=(double**)multialloc(sizeof(double),2,2*N,2*M);
      for (i=0;i<2*N;i++)
	for (j=0;j<2*M;j++)
	  tmp[i][j] = data->X[i/2][j/2];
      write_mat_double_image(tmp,2*N,2*M,matrix,fout);
      multifree((void**)tmp,2);
    }

  sprintf(matrix,"sigma");
  write_mat_double_image(sigma,1,1,matrix,fout);

  sprintf(matrix,"fov");
  fov=geom->dia/10.0;
  ptmp=&fov;
  write_mat_double_image(&ptmp,1,1,matrix,fout);

  matClose(fout);
}
  


void init_X_to_constant(struct all_data *data,struct geom_info *geom,double **profiles,struct column *sp_col)
{
  double init_const;
  int i,j;

  init_const=calc_init_const(data,geom,profiles,sp_col);
  
  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_x;j++)
      if (data->recon_mask[i][j])
	data->X[i][j]=init_const;
      else
	data->X[i][j]=0.0;
}



double calc_init_const(struct all_data *data,struct geom_info *geom,double **profiles,struct column *sp_col)
{
  double sumA=0.0,sumY=0.0;
  int i,j,t;
  double init_const;
  
  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_x;j++)
      if (data->recon_mask[i][j])
	{
	  A_comp(i,j,&data->sino,geom,profiles,sp_col,0);
	  for (t=0;t<sp_col->n_row;t++)
	    sumA = sumA + sp_col->val[t];
	}
  
  for (i=0;i<data->sino.n_theta;i++)
    for (j=0;j<data->sino.n_t;j++)
      sumY = sumY + data->Y[i][j];
  
  /*printf("sumY=%f sumA=%f\n",sumY,sumA);*/
  
  init_const = sumY/sumA;

  printf("Initializing to constant = %g\n\n",init_const);

  return(init_const);
}
  

void calculate_pix_recon_mask(struct all_data *data,struct geom_info *geom)
{
  int i,j;
  double r_sq,x,y,r_sq_max;
  
  r_sq_max = data->sino.t_0 * data->sino.t_0 ;
  x=y=r_sq=r_sq_max=0.0;


  data->recon_mask=(int**)multialloc(sizeof(int),2,geom->n_y,geom->n_x);
  
  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_x;j++)
      data->recon_mask[i][j]=1.0;

#ifdef RECON_MASK
  r_sq_max = data->sino.t_0 * data->sino.t_0 ;
  


  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_x;j++)
      {
	y = geom->y_0 + i*geom->delta;
	x = geom->x_0 + j*geom->delta;
	
	r_sq=x*x+y*y;
	
	if (r_sq>r_sq_max)
	  data->recon_mask[i][j]=0.0;
      }
#endif
}
  

void clip_initial(struct all_data *data,struct geom_info *geom)
{
  int i,j;
  
  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_x;j++)
      if (data->recon_mask[i][j]==0)
	data->X[i][j]=0.0;
}


 
void free_memory(struct all_data *data, struct column *sp_col,double **sigma,struct cmdline_params *cmdline)
{
  printf("Freeing memory ...\n");
  
  free_img((void **)data->X);
  free_img((void **)sigma);
  free_img((void **)data->Y);
  free((void *)sp_col->val);
  free((void *)sp_col->row);
  free_img((void **)data->recon_mask);

  if (cmdline->mask_FLAG)
    free_img((void **)data->mask);
  if (cmdline->Transcor_FLAG)
    free_img((void **)sp_col->A_cor.Y_t);

}
